import numpy as np
from matplotlib import pyplot as plt
from matplotlib import patches, ticker
import matplotlib.image as mpimg
from glob import glob
from scipy.interpolate import interp2d, bisplev, interp1d, UnivariateSpline
from scipy.optimize import curve_fit
import re
from glob import glob
from utils import *

cm = 1e7

data_root = "../Data/"

layer1 = data_root + 'single31x31_mesh5_eps39_z102.25.csv'
layer2 = data_root + 'single31x31_mesh5_eps39_z101.75.csv'
fname = data_root + 'single31x31_mesh5_eps39_z102.25.csv'

def z_der_fun_two(layer1,layer2,dz = 0.2):
    data1 = load_data(layer1)
    xlabels, ylabels = get_xy(data1)
    data_map1 = get_interpolated_map(get_data_map(data1),xlabels,ylabels, return_func=True)

    data2 = load_data(layer2)
    xlabels, ylabels = get_xy(data2)
    data_map2 = get_interpolated_map(get_data_map(data2),xlabels,ylabels, return_func=True)

    return lambda x,y: (data_map2[2](x,y)-data_map1[2](x,y))/dz


def generate_derivatives(data):
    #Here I compute the derrivatibes
    xlabels, ylabels = get_xy(data)
    x_dim = xlabels[-1]-xlabels[0]    #size in x-direction in nm
    y_dim = ylabels[-1]-ylabels[0]    #size in y-directions in nm
    single_data_mapL = get_data_map(data)

    interpolated_dx = get_interpolated_map(single_data_mapL, xlabels, ylabels, dx=1, 
                                                       return_func=True)[2]
    interpolated_dy = get_interpolated_map(single_data_mapL, xlabels, ylabels, dy=1, 
                                                       return_func=True)[2]
                                                       
    dz = abs(get_z_from_path(layer1) - get_z_from_path(layer2))
    print(f"dz = {dz}")
    interpolated_dz = z_der_fun_two(layer1, layer2, dz=dz)
    #interpolated_dz = get_interpolated_map(der_z_map_f(100.2, dz=1), xlabels, ylabels,return_func=True)[2]

    derr = [interpolated_dx,interpolated_dy, interpolated_dz]  #[dE/dx, dE/dy]
    return x_dim, y_dim, derr

def generate_random_positions(number_of_defects, x_dim, y_dim, z_dim=None):
    '''
    Function generates random positions of defects in 2D or 3D
    returns array (3,N) or (2,N)
    '''
    x_range = x_dim[-1]-x_dim[0]
    y_range = y_dim[-1]-y_dim[0]

    
    if z_dim:
        return np.array([x_dim[0]+x_range*np.random.uniform(size = number_of_defects),
                         y_dim[0]+y_xrange*np.random.uniform(size = number_of_defects),
                         z_dim[0]+z_xrange*np.random.uniform(size = number_of_defects)]).T
    else:
        return np.array([x_dim[0]+x_range*np.random.uniform(size = number_of_defects),
                         y_dim[0]+y_range*np.random.uniform(size = number_of_defects)]).T



def generate_random_shifts(number_of_defects, x_dim, y_dim, sigma_r, z_dim=None):
    '''
    Function moves each of the defect by random variable, such that the distance jumped has std=sigma_r.
    returns array (3,N) or (2,N)
    '''
    # Normalization experimental, still don't know how to get that analyitically
    if z_dim:
        return np.array([sigma_r/np.sqrt(3)*np.random.normal(size = number_of_defects),
                         sigma_r/np.sqrt(3)*np.random.normal(size = number_of_defects),
                         sigma_r/np.sqrt(3)*np.random.normal(size = number_of_defects)]).T/0.9213806971076426
    else:
        return np.array([sigma_r/np.sqrt(2)*np.random.normal(size = number_of_defects),
                         sigma_r/np.sqrt(2)*np.random.normal(size = number_of_defects)]).T/0.8865204310875154
    

def generate_random_positions(number_of_defects, x_dim, y_dim, z_dim=None):
    '''
    Function generates random positions of defects in 2D or 3D
    returns array (3,N) or (2,N)
    '''
    x_range = x_dim[-1]-x_dim[0]
    y_range = y_dim[-1]-y_dim[0]

    
    if z_dim:
        return np.array([x_dim[0]+x_range*np.random.uniform(size = number_of_defects),
                         y_dim[0]+y_xrange*np.random.uniform(size = number_of_defects),
                         z_dim[0]+z_xrange*np.random.uniform(size = number_of_defects)]).T
    else:
        return np.array([x_dim[0]+x_range*np.random.uniform(size = number_of_defects),
                         y_dim[0]+y_range*np.random.uniform(size = number_of_defects)]).T

def generate_random_shifts(number_of_defects, x_dim, y_dim, sigma_r, z_dim=None):
    '''
    Function moves each of the defect by random variable, such that the distance jumped has std=sigma_r.
    returns array (3,N) or (2,N)
    '''
    # Normalization experimental, still don't know how to get that analyitically
    if z_dim:
        return np.array([sigma_r/np.sqrt(3)*np.random.normal(size = number_of_defects),
                         sigma_r/np.sqrt(3)*np.random.normal(size = number_of_defects),
                         sigma_r/np.sqrt(3)*np.random.normal(size = number_of_defects)]).T/0.9213806971076426
    else:
        return np.array([sigma_r/np.sqrt(2)*np.random.normal(size = number_of_defects),
                         sigma_r/np.sqrt(2)*np.random.normal(size = number_of_defects)]).T/0.8865204310875154
    
def generate_etas(x_dim,y_dim,derr, rho, sigma_r):
    charge_density = rho/cm**2
    number_of_defects = int(x_dim*y_dim*charge_density)
    r = generate_random_positions(number_of_defects, x_dim, y_dim, z_dim=100)  #wylosuj położenia defektów
    dr = generate_random_shifts(number_of_defects, x_dim, y_dim, z_dim=True, sigma_r = sigma_r)
    etasL = [np.sum([dr[n,k]*derr[k](r[n,0],r[n,1]) for k in range(3)]) for n in range(number_of_defects)]
    return etasL


def generate_etas_fpd(x_dim,y_dim, derr, rho, sigma_r, om_range, om_range_ext, z_jump = True):
    
    charge_density = rho/cm**2
    x_range = x_dim[-1]-x_dim[0]
    y_range = y_dim[-1]-y_dim[0]
    m = 1
    number_of_defects = int(m**2*x_range*y_range*charge_density)
    
    center = [(x_dim[-1]+x_dim[0])/2,(y_dim[-1]+y_dim[0])/2]
    fpd = number_of_defects/np.log10(om_range[1]/om_range[0])
    number_of_defects = int(np.log10(om_range_ext[1]/om_range_ext[0])*fpd)
    r = generate_random_positions(number_of_defects, center[0]+m*(x_dim-center[0]), 
                                  center[1]+m*(y_dim-center[1]), z_dim=False)  #wylosuj położenia defektów
    dr = generate_random_shifts(number_of_defects, x_dim, y_dim, z_dim=z_jump, sigma_r = sigma_r)
    if z_jump:
        etasL = [np.sum([dr[n,k]*derr[k](r[n,0],r[n,1]) for k in range(3)]) for n in range(number_of_defects)]
    else:
        etasL = [np.sum([dr[n,k]*derr[k](r[n,0],r[n,1]) for k in range(2)]) for n in range(number_of_defects)]
    return etasL
    

### Plot preparation

def prepare_plot(figsize, height_ratios=[1,1,1]):
    fig = plt.figure(1, figsize=figsize)

    gs = fig.add_gridspec(3,1, hspace=0, wspace=0, height_ratios=height_ratios, width_ratios=[1])
    # gs.tight_layout(fig, pad=0, h_pad=0, w_pad=0)
    ax1 = fig.add_subplot(gs[0, 0])
    ax2 = fig.add_subplot(gs[1, 0], sharex=ax1)
    return [ax1, ax2]


def get_coeffs(sigma):
    plt.sca(axes[1])
    # plt.figure(figsize=(4,3), facecolor = "w")
    xses = drs
    ases = []
    bs = []
    for k in range(len(sigma[0,:,0])):
        # args = np.argwhere((V > 0) & (V <= 1.5)).reshape(-1)
        xspace = np.arange(len(drs))+1
        yvals =  np.average(np.sqrt(sigma[:,k,:]), axis=0)
        f = lambda x, a, b: a*x + b
        popt, _ = curve_fit(f, drs, yvals)
        ases.append(popt[0])
        bs.append(popt[1])
    return ases,bs


def plot_distributions(axes,sigma, color_dashed = "k"):
    plt.sca(axes)
    plt.subplots_adjust(left=0.18, bottom=0.16, right=0.97, top=0.94)
    xses = drs
    color = ["b","orange","g","r","purple","darkred"]
    labels = [r"$5\times 10^{9}$",r"$10^{10}$",r"$2\times 10^{10}$","",""]  
    for k in range(3):
        plt.violinplot(np.sqrt(sigma[:,k,:]),positions = np.arange(len(xses))+0.80+k*0.2)
        #plt.violinplot(np.sqrt(etas2R[:,k,:]),positions = np.arange(len(xses))+0.8+k*0.15)
        plt.plot([],[],color = color[k],alpha=0.5,lw=3,label=labels[k])
    #plt.yscale("log")
    l = plt.legend(title=r"$\rho$ [cm$^{-2}$]",
                   loc="lower right", ncol=1, frameon=True, fontsize=8.5, framealpha=0.4)
    #plt.ylim(bottom=1e-4, top=8*1e2)
    plt.xticks(np.arange(len(xses))+1,np.round(xses,2))
    #plt.plot([-5,len(xses)+1],[1,1], "g")iii
    #plt.plot([-5,len(xses)+1],[10,10], "r")
    plt.plot([-5,len(xses)+1],[8,8], "--",color=color_dashed)

    plt.xlabel(r"Displacemt size dr (nm)")
    # plt.ylabel(r"Noise amplitude $\sigma\,(\mu$eV)")
    plt.ylabel(r"$\sigma_\mathrm{spin}$ (neV)")
    plt.xlim(0.38,len(xses)+0.5)
    #plt.ylim(-0.1,25)
    plt.grid(lw=0.5)
    plt.grid(True)
    plt.yscale("log")
    #plt.ylim(1e-2,1e1)
    #plt.text(0.5,3.2*1e-1,"a)")
    axes.tick_params(labelbottom=True, direction="in")
    #ax =  plt.twinx()
    #ax.set_ylabel("Charge qubit $T_2^*$ (ps)")
    #ax.set_ylim(0,30)
    #ax.set_yticks([np.sqrt(2)*hbar/0.05,np.sqrt(2)*hbar/0.1,np.sqrt(2)*hbar/0.5])
    #ax.set_yticklabels([50, 100, 500])
    #plt.title(r"Density="+str(charge_densities[inda]/1e9)+" (1e9cm$^{-2})$")




data = load_data(fname)
xlabels, ylabels = get_xy(data)
data_map = get_data_map(data)

charge_densities = [5*1e9,1e10,2*1e10,5*1e10,1e11]
#charge_densities = [5*1e9,5*1e10,5*1e11,5*1e9*np.sqrt(10),5*1e10*np.sqrt(10)]    # Density of charges
drs = np.linspace(0.1,0.9,5)      # Average jump size

trials = 500
etas_z = []
etas_xy = []
params = []
range_om = (1e-6,1e6)
range_om_ext = (1e-6,1e6)
x_dim,y_dim,derr = generate_derivatives(data)
np.random.seed(61)
for tr in range(trials):
    if tr %10 ==0: print(tr)
    for rn,rho in enumerate(charge_densities):
        for drn,dr in enumerate(drs):
            etasL = generate_etas_fpd(xlabels, ylabels, derr, rho, dr, range_om, range_om_ext, z_jump=True)
            etas_z.append(np.array([etasL]))
            params.append([tr,rn,drn])
            
            etasL = generate_etas_fpd(xlabels, ylabels, derr, rho, dr, range_om, range_om_ext, z_jump=False)
            etas_xy.append(np.array([etasL]))
            #params.append([tr,rn,drn])
            

sigma_iso = np.zeros((trials,len(charge_densities),len(drs)))
sigma_xy = np.zeros((trials,len(charge_densities),len(drs)))
for k,etas0 in enumerate(etas_z):
    sigma_iso[params[k][0],params[k][1],params[k][2]] = np.sum(etas0[0]**2)
  
for k,etas0 in enumerate(etas_xy):
    sigma_xy[params[k][0],params[k][1],params[k][2]] = np.sum(etas0[0]**2)
    
    
    
hbar = 0.658



axes = prepare_plot(figsize=(3, 8))
plot_distributions(axes[0], sigma_iso)
axes[0].text(0.8,6,"Isotropic model")
plot_distributions(axes[1], sigma_xy, color_dashed = "red")
axes[1].text(0.8,6,"Planar model")
a_avg_iso,b_iso = get_coeffs(sigma_iso)
a_avg_xy,b_xy = get_coeffs(sigma_xy)



plt.figure(figsize=(3,1.8))
plt.subplots_adjust(left=0.18, bottom=0.16, right=0.97, top=0.94)
charges = np.array(charge_densities, dtype=float)
plt.yscale('log')
plt.xscale('log')
plt.xlabel(r"charge density $\rho$ (cm$^{-2}$)")
plt.ylabel(r"$\partial \langle \sigma \rangle/\partial\delta r (n$eV$/\mathrm{nm})$")

xspace = np.linspace(min(charges), max(charges), 100)
f = lambda x, a, b: a*x + b
popt, _ = curve_fit(f, np.log(charges), np.log(a_avg_iso))
print("log(a_avg(rho)) = %.3f * log(rho)"%popt[0])
plt.plot(xspace, np.exp(f(np.log(xspace), *popt)), '--', label=r"$\gamma_\mathrm{iso}=%.3f"%popt[0]+"$", color="black")
plt.plot(charges, a_avg_iso, 'o', color="black")
gamma_iso = popt[0]

popt, _ = curve_fit(f, np.log(charges), np.log(a_avg_xy ))
print("log(a_std(rho)) = %.3f * log(rho)"%popt[0])
plt.plot(xspace, np.exp(f(np.log(xspace), *popt)), '--', label=r"$\gamma_\mathrm{xy}=%.3f"%popt[0]+"$", color="red")
plt.plot(charges, a_avg_xy, 'o', color="red")
plt.legend(fontsize=10, frameon=True)
plt.grid("True")
#plt.ylim(6*1e-1,2*1e1)
gamma_xy = popt[0]
#plt.ylim(bottom=0.8*1e1)
#plt.text(5.2*1e9,1.15*1e2,"a)")


charges = np.array(charge_densities, dtype=float)
xspace = np.linspace(min(charges), max(charges), 100)
f = lambda x, a,b: a*x**(gamma_xy)+b
popt,_= curve_fit(f, charges, a_avg_xy)
arho_xy,brho_xy = popt


f = lambda x, a,b: a*x**(gamma_iso)+b
popt,_= curve_fit(f, charges, a_avg_iso)
arho_iso,brho_iso = popt

